"""
This module contains functions for calculating various metrics.
"""
# pylint: disable=too-many-arguments
# pylint: disable=unused-argument
# pylint: disable=too-many-locals

from typing import Iterable

import pandas as pd

from bssir.calculator import Calculator
from bssir.calculator.quantile import _EquivalenceScale, _QuantileBase

from .api import api


calculate = Calculator(api)


def __get_optional_params(local_variables: dict) -> dict:
    return {key: value for key, value in local_variables.items() if value is not None}


def weighted_average(
    table: pd.DataFrame,
    weight_col: str | None = None,
    columns: list[str] | None = None,
) -> pd.Series:
    """Calculates the weighted average of the given columns.

    Parameters
    ----------
    table : pd.DataFrame
        The table to calculate the weighted average for.
    weight_col : str | None, optional
        The column to use as the weight, by default None
    columns : list[str] | None, optional
        The columns to calculate the weighted average for, by default None

    Returns
    -------
    pd.Series
        The weighted average of the given columns.
    """
    parameters = __get_optional_params(locals())
    return calculate.weighted_average(**parameters)


def average_table(
    table: pd.DataFrame,
    weight_col: str | None = None,
    groupby: list[str] | None = None,
    columns: list[str] | None = None,
) -> pd.DataFrame:
    """Calculates the weighted average of the given columns for each group.

    Parameters
    ----------
    table : pd.DataFrame
        The table to calculate the weighted average for.
    weight_col : str | None, optional
        The column to use as the weight, by default None
    groupby : list[str] | None, optional
        The columns to group by, by default None
    columns : list[str] | None, optional
        The columns to calculate the weighted average for, by default None

    Returns
    -------
    pd.DataFrame
        The weighted average of the given columns for each group.
    """
    parameters = __get_optional_params(locals())
    return calculate.average_table(**parameters)


def quantile(
    table: pd.DataFrame | pd.Series | None = None,
    quantile_column_name: str | None = None,
    bins: int = -1,
    weight_column: str | None = None,
    on_variable: _QuantileBase | None = None,
    on_column: str | None = None,
    weighted: bool = True,
    adjust_weight_for_household_size: bool = False,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: str | Iterable[str] | None = None,
    years: list[int] | None = None,
) -> pd.Series:
    """Calculates weighted quantiles for the given column.

    Parameters
    ----------
    table : pd.DataFrame | pd.Series, optional
        Table containing column to calculate quantiles for.
    on_variable : _QuantileBase, optional
        Variable to calculate quantiles on.
    on_column : str, optional
        Column name to calculate quantiles on.
    bins : int, optional
        Number of quantile bins to create.

    Other parameters
    ----------------
    quantile_column_name : str, optional
        Name for added quantile column.
    weight_column : str, optional
        Column name to use for weighting rows.
    weighted : bool, optional
        Whether to calculate weighted quantiles.
    adjust_weight_for_household_size : bool, optional
        Adjust weights for household size.
    equivalence_scale : _EquivalenceScale, optional
        Household equivalence scale to use.
    for_all : bool, optional
        Calculate quantiles across all rows.
    annual : bool, optional
        Annualize quantiles.
    groupby : str | list[str], optional
        Column(s) to group by.
    years : list[int], optional
        Years to include.

    Returns
    -------
    pd.Series
        Series with quantiles for each row.
    """
    parameters = __get_optional_params(locals())
    return calculate.quantile(**parameters)


def add_quantile(
    table: pd.DataFrame | pd.Series | None = None,
    quantile_column_name: str | None = None,
    bins: int = -1,
    weight_column: str | None = None,
    on_variable: _QuantileBase | None = None,
    on_column: str | None = None,
    weighted: bool = True,
    adjust_weight_for_household_size: bool = False,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: str | Iterable[str] | None = None,
    years: list[int] | None = None,
) -> pd.DataFrame:
    """Adds quantiles as a column to the input table.

    Parameters
    ----------
    table : DataFrame | Series, optional
        Table to add quantiles to.
    bins : int, optional
        Number of quantile bins to create.
    on_variable : _QuantileBase, optional
        Variable to calculate quantiles on.
    on_column : str, optional
        Column to calculate quantiles on.

    Other parameters
    ----------------
    quantile_column_name : str, optional
        Name for added quantile column.
    weight_column : str, optional
        Column name to use for weighting rows.
    weighted : bool, optional
        Whether to calculate weighted quantiles.
    adjust_weight_for_household_size : bool, optional
        Adjust weights for household size.
    equivalence_scale : _EquivalenceScale, optional
        Household equivalence scale to use.
    for_all : bool, optional
        Calculate quantiles across all rows.
    annual : bool, optional
        Annualize quantiles.
    groupby : str | list[str], optional
        Column(s) to group by.
    years : list[int], optional
        Years to include.

    Returns
    -------
    DataFrame
        Input table with added quantile column.
    """
    parameters = __get_optional_params(locals())
    return calculate.add_quantile(**parameters)


def add_decile(
    table: pd.DataFrame | pd.Series | None = None,
    quantile_column_name: str | None = None,
    weight_column: str | None = None,
    on_variable: _QuantileBase | None = None,
    on_column: str | None = None,
    weighted: bool = True,
    adjust_weight_for_household_size: bool = False,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: str | Iterable[str] | None = None,
    years: list[int] | None = None,
) -> pd.DataFrame:
    """Adds deciles as a column to the input table.

    Parameters
    ----------
    table : DataFrame | Series, optional
        Table to add quantiles to.
    bins : int, optional
        Number of quantile bins to create.
    on_variable : _QuantileBase, optional
        Variable to calculate quantiles on.
    on_column : str, optional
        Column to calculate quantiles on.

    Other parameters
    ----------------
    quantile_column_name : str, optional
        Name for added quantile column.
    weight_column : str, optional
        Column name to use for weighting rows.
    weighted : bool, optional
        Whether to calculate weighted quantiles.
    adjust_weight_for_household_size : bool, optional
        Adjust weights for household size.
    equivalence_scale : _EquivalenceScale, optional
        Household equivalence scale to use.
    for_all : bool, optional
        Calculate quantiles across all rows.
    annual : bool, optional
        Annualize quantiles.
    groupby : str | list[str], optional
        Column(s) to group by.
    years : list[int], optional
        Years to include.

    Returns
    -------
    DataFrame
        Input table with added decile column.
    """
    parameters = __get_optional_params(locals())
    return calculate.add_decile(**parameters)


def add_percentile(
    table: pd.DataFrame | pd.Series | None = None,
    quantile_column_name: str | None = None,
    weight_column: str | None = None,
    on_variable: _QuantileBase | None = None,
    on_column: str | None = None,
    weighted: bool = True,
    adjust_weight_for_household_size: bool = False,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: str | Iterable[str] | None = None,
    years: list[int] | None = None,
) -> pd.DataFrame:
    """Adds percentiles as a column to the input table.

    Parameters
    ----------
    table : DataFrame | Series, optional
        Table to add quantiles to.
    bins : int, optional
        Number of quantile bins to create.
    on_variable : _QuantileBase, optional
        Variable to calculate quantiles on.
    on_column : str, optional
        Column to calculate quantiles on.

    Other parameters
    ----------------
    quantile_column_name : str, optional
        Name for added quantile column.
    weight_column : str, optional
        Column name to use for weighting rows.
    weighted : bool, optional
        Whether to calculate weighted quantiles.
    adjust_weight_for_household_size : bool, optional
        Adjust weights for household size.
    equivalence_scale : _EquivalenceScale, optional
        Household equivalence scale to use.
    for_all : bool, optional
        Calculate quantiles across all rows.
    annual : bool, optional
        Annualize quantiles.
    groupby : str | list[str], optional
        Column(s) to group by.
    years : list[int], optional
        Years to include.

    Returns
    -------
    DataFrame
        Input table with added percentile column.
    """
    parameters = __get_optional_params(locals())
    return calculate.add_percentile(**parameters)



from typing import Literal

import pandas as pd

from ...api import api
from ...calculate import calculate

from .common import create_expenditure_tiers, filter_urban_rural


def create_table_for_a_year(
    year: int, urban_rural: Literal["urban", "rural"]
) -> pd.DataFrame:
    literate_members = (
        api.load_table("members_properties", years=year)
        .assign(Employed=lambda df: df["Activity_Status"] == "Employed")
        .groupby(["Year", "ID"])["Employed"]
        .sum()
        .rename("Value")
    )

    have_income_members = (
        api.load_table("Members_Total_Income", years=year)
        .groupby(["Year", "ID"])["Member_Number"]
        .count()
        .rename("Have_Income")
    )

    members = (
        api.load_table("Number_of_Members", years=year)
        .set_index(["Year", "ID"])
        .loc[:, "Members"]
    )

    table = pd.concat(
        [
            members,
            have_income_members,
            literate_members,
            create_expenditure_tiers(urban_rural=urban_rural, year=year),
            api.load_table("Weight", years=year).set_index(["Year", "ID"])["Weight"],
        ],
        axis="columns",
    )
    table = filter_urban_rural(table, urban_rural)

    total = calculate.weighted_average(
        table, columns=["Members", "Have_Income", "Value"]
    )

    average_number_of_members = (
        table.groupby("Expenditure_Tier", observed=True)
        .apply(calculate.weighted_average, columns=["Members", "Have_Income", "Value"])
        .rename(
            columns={
                "Members": "Average_Members",
                "Have_Income": "Average_Have_Income",
                "Value": "Average_Employed",
            }
        )
    )
    average_number_of_members.loc["Total"] = total.to_list()

    table["Value"] = table["Value"].clip(0, 5)
    weight_table = (
        table.groupby(["Value", "Expenditure_Tier"], observed=True)["Weight"]
        .sum()
        .unstack()
    )
    index_weight_sum = weight_table.sum(axis="index")
    columns_weight_sum = weight_table.sum(axis="columns")
    col_sum = columns_weight_sum.div(columns_weight_sum.sum()).mul(100).rename("Total")
    main_table = weight_table.div(index_weight_sum).mul(100)
    main_table = pd.concat([col_sum, main_table], axis="columns")

    average_number_of_members = average_number_of_members.T[main_table.columns]

    main_table = pd.concat(
        [
            average_number_of_members,
            main_table.sum().to_frame("Total").T,
            main_table,
        ]
    )

    main_table = main_table.fillna(0)
    main_table.index = main_table.index.astype(str)

    return main_table


def main(years: list[int], urban_rural: Literal["urban", "rural"]):
    if len(years) == 1:
        return create_table_for_a_year(years[0], urban_rural)
    raise ValueError



from typing import Literal

import pandas as pd

from ...api import api
from ...calculate import calculate

from .common import filter_urban_rural


def create_table_for_a_year(
    year: int, urban_rural: Literal["urban", "rural", "both"]
) -> pd.DataFrame:
    members = (
        api.load_table("Number_of_Members", years=year)
        .set_index(["Year", "ID"])
        .loc[:, "Members"]
        .rename("Value")
    )

    table = pd.concat(
        [
            members,
            calculate.quantile(
                years=[year],
                on_variable="Net_Expenditure",
                groupby=["Urban_Rural"],
                bins=10,
                quantile_column_name="Decile",
            ),
            api.load_table("Weight", years=year).set_index(["Year", "ID"])["Weight"],
        ],
        axis="columns",
    )
    table = filter_urban_rural(table, urban_rural)

    total = calculate.weighted_average(table, columns=["Value"]).iloc[0]

    average_number_of_members = (
        table.groupby("Decile", observed=True)
        .apply(calculate.weighted_average, columns=["Value"])
        .rename(columns={"Value": "Average"})
    )
    average_number_of_members.loc["Total"] = total

    table["Value"] = table["Value"].clip(0, 10)
    weight_table = (
        table.groupby(["Value", "Decile"], observed=True)["Weight"].sum().unstack()
    )
    index_weight_sum = weight_table.sum(axis="index")
    columns_weight_sum = weight_table.sum(axis="columns")
    col_sum = columns_weight_sum.div(columns_weight_sum.sum()).mul(100).rename("Total")
    main_table = weight_table.div(index_weight_sum).mul(100)
    main_table = pd.concat([col_sum, main_table], axis="columns")

    average_number_of_members = average_number_of_members.T[main_table.columns]

    main_table = pd.concat(
        [
            average_number_of_members,
            main_table.sum().to_frame("Total").T,
            main_table,
        ]
    )

    main_table = main_table.fillna(0)
    main_table.index = main_table.index.astype(str)
    main_table.columns = main_table.columns.astype(str)

    return main_table


def main(years: list[int], urban_rural: Literal["urban", "rural"]):
    if len(years) == 1:
        return create_table_for_a_year(years[0], urban_rural)
    raise ValueError



from typing import Literal

import pandas as pd

from ...api import api

from .common import create_expenditure_tiers, filter_urban_rural

INDEX = [
    "Total",
    "Agriculture, forestry and fishing",  #
    "Mining and quarrying",
    "Manufacturing",
    "Electricity, gas, steam and air conditioning supply",
    "Construction",
    "Wholesale and retail trade; repair of motor vehicles and motorcycles",
    "Accommodation and food service activities",
    "Transportation and storage",
    "Information and communication",
    "Financial and insurance activities",
    "Real estate activities",
    "Public administration and defence; compulsory social security",
    "Education",
    "Human health and social work activities",
    "Other service activities",
    "Administrative and support service activities",
    "Arts, entertainment and recreation",
    "Professional, scientific and technical activities",
    "Water supply; sewerage, waste management and remediation activities",
    (
        "Activities of households as employers; undifferentiated goods- "
        "and services-producing activities of households for own use"
    ),
    "Unemployed",
]


def create_table_for_a_year(
    year: int, urban_rural: Literal["urban", "rural"]
) -> pd.DataFrame:
    activity_state = (
        api.load_table("members_properties", years=1401)
        .set_index(["Year", "ID", "Member_Number"])
        .loc[lambda df: df["Relationship"] == "Head", ["Activity_State"]]
    )

    income_table = pd.concat(
        [
            api.load_table("employment_income", years=year),
            api.load_table("self_employed_income", years=year),
        ],
        ignore_index=True,
    )
    industry = (
        income_table.drop_duplicates(["ID", "Member_Number"], keep="first")
        .pipe(api.add_classification, target="Industry_Code", aspects="title")
        .set_index(["Year", "ID", "Member_Number"])
        .loc[:, ["Industry"]]
    )

    industry = activity_state.join(industry, how="left").droplevel(-1)
    filt = industry["Activity_State"] != "Employed"
    industry.loc[filt, "Industry"] = "Unemployed"
    filt = industry["Industry"].isna() & (industry["Activity_State"] == "Employed")
    industry.loc[filt, "Industry"] = "Unknown or Other"

    industry = industry["Industry"]

    table = pd.concat(
        [
            industry,
            create_expenditure_tiers(urban_rural=urban_rural, year=year),
            api.load_table("Weight", years=year).set_index(["Year", "ID"])["Weight"],
        ],
        axis="columns",
    )

    table = filter_urban_rural(table, urban_rural)

    weight_table = (
        table.groupby(["Industry", "Expenditure_Tier"], observed=True)["Weight"]
        .sum()
        .unstack()
    )
    index_weight_sum = weight_table.sum(axis="index")
    columns_weight_sum = weight_table.sum(axis="columns")
    col_sum = columns_weight_sum.div(columns_weight_sum.sum()).mul(100).rename("Total")
    main_table = weight_table.div(index_weight_sum).mul(100)
    main_table = pd.concat([col_sum, main_table], axis="columns")
    main_table = pd.concat(
        [
            main_table.sum().to_frame("Total").T,
            main_table,
        ]
    )

    main_table = main_table.fillna(0)
    main_table = main_table.loc[INDEX]
    return main_table


def main(years: list[int], urban_rural: Literal["urban", "rural"]):
    if len(years) == 1:
        return create_table_for_a_year(years[0], urban_rural)
    raise ValueError



from pathlib import Path
from typing import Optional

from bssir.metadata_reader import config, _Years
from bssir.api import API
from bssir.maintainer import Maintainer


defaults, metadata = config.set_package_config(Path(__file__).parent)
api = API(defaults=defaults, metadata=metadata)

def create_and_upload(
    years: _Years = "all",
    table_names: Optional[list[str]] = None,
    mirror_name: Optional[str] = None,
    recreate_files: bool = True
) -> None:
    if recreate_files:
        api.setup(
            years,
            table_names=table_names,
            method="create_from_raw",
            download_source="original",
        )
    maintainer = Maintainer(
        lib_defaults=defaults,
        lib_metadata=metadata,
        mirror_name=mirror_name,
    )
    years = api.utils.parse_years(years)
    maintainer.upload_raw_files(years=years)
    maintainer.upload_cleaned_files(years=years, table_names=table_names)
